/***
 * Juliac
 * Copyright (C) 2009-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.compile.jdk6;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import javax.tools.Diagnostic;
import javax.tools.Diagnostic.Kind;
import javax.tools.DiagnosticCollector;
import javax.tools.JavaCompiler;
import javax.tools.JavaCompiler.CompilationTask;
import javax.tools.JavaFileManager;
import javax.tools.JavaFileObject;
import javax.tools.StandardJavaFileManager;
import javax.tools.ToolProvider;

import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.SourceFile;
import org.objectweb.fractal.juliac.Utils;
import org.objectweb.fractal.juliac.api.CompileSupportItf;
import org.objectweb.fractal.juliac.conf.JDKLevel;
import org.objectweb.fractal.juliac.conf.JuliacConfig;

/**
 * This class implements a compilation service based on the Java Compilation API
 * (JSR 199) provided by JDK 6.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 2.2.3
 */
public class CompileSupportImpl implements CompileSupportItf {

    // -----------------------------------------------------------------------
    // Implementation of the JuliacModuleItf interface
    // -----------------------------------------------------------------------

	final public static Class<CompileSupportItf> SERVICE_TYPE =
		CompileSupportItf.class;
	
	private Juliac jc;
	
    public void init( Juliac jc ) {
    	this.jc = jc;
        JuliacConfig jconf = jc.getJuliacConfig();
        jconf.register(SERVICE_TYPE,this);
    }
    
    public void close( Juliac jc ) {
        JuliacConfig jconf = jc.getJuliacConfig();
        jconf.unregister(SERVICE_TYPE,this);
    }

    
    // -----------------------------------------------------------------------
    // Implementation of the CompileSupportItf interface
    // -----------------------------------------------------------------------

    /**
     * Compile the specified source files and dump the resulting bytecode files
     * into the specified directory.
     * 
     * @param sfs       the source files
     * @param classDir  the directory for storing bytecode files
     * @return 
     * 		the paths relative to <code>classDir</code> of compiled bytecode
     * 		files 
     * @throws IOException  if an error occurs during the compilation
     */
    public List<String> compile( List<SourceFile> sfs, File classDir )
    throws IOException {
        
        JuliacConfig jconf = jc.getJuliacConfig();
    	Logger logger = jconf.getCompilationLogger();
        JavaCompiler compiler = ToolProvider.getSystemJavaCompiler();

        /*
         * Prepare the source files.
         */
        List<JavaFileObject> jfos = new ArrayList<JavaFileObject>();
        for (SourceFile sf : sfs) {
            char[] content = sf.getContent();
            String fileName = sf.getName();
            JavaFileObject jfo = new JavaSourceFromString(fileName,content);
            jfos.add(jfo);
        }
        
        /*
         * Prepare compiler options.
         */
        List<String> options = new ArrayList<String>();

        // Source and target compatibility levels
        JDKLevel sourceLevel = jconf.getSourceLevel();
        JDKLevel targetLevel = jconf.getTargetLevel();
        options.add("-source");
        options.add(sourceLevel.getVersion());
        options.add("-target");
        options.add(targetLevel.getVersion());
        
        // Add the classpath option
        options.add("-classpath");
        List<String> cpes = getClassPathEntries();
        StringBuffer sb = new StringBuffer();
        boolean first = true;
        for (String cpe : cpes) {
            if(first) {first=false;} else {sb.append(File.pathSeparatorChar);}
            sb.append(cpe);
        }
        options.add(sb.toString());
        
        // Directory for .class files
        String path = classDir.getAbsolutePath();
        options.add("-d");
        options.add(path);
        
        /*
         * Prepare a Java file manager for retrieving the names of the created
         * .class files.
         */
        StandardJavaFileManager stdFileManager =
            compiler.getStandardFileManager(null,null,null);
        JavaFileManagerLogger<JavaFileManager> fileManager =
            new JavaFileManagerLogger<JavaFileManager>(stdFileManager);

        /*
         * Compile the code.
         */
        DiagnosticCollector<JavaFileObject> dc =
            new DiagnosticCollector<JavaFileObject>();
        CompilationTask ct =
            compiler.getTask(null,fileManager,dc,options,null,jfos);
        boolean success = ct.call();
        
        /*
         * Report errors, warnings and info messages if any.
         */
        List<Diagnostic<? extends JavaFileObject>> diags = dc.getDiagnostics();
        for (Diagnostic<?> diag : diags) {
        	Kind kind = diag.getKind();
        	if( kind.equals(Kind.ERROR) ) {
                logger.severe(diag.toString());
        	}
        	else if( kind.equals(Kind.WARNING) ) {
                logger.warning(diag.toString());
        	}
        	else {
        		logger.info(diag.toString());
        	}
        }
        if (!success) {
            String msg = "Compilation error";
            throw new IOException(msg);
        }

        /*
         * Retrieve output file names.
         */
        List<String> classNames = fileManager.getClassNames();
        List<String> outputFileNames = new ArrayList<String>();
        for (String className : classNames) {
            String outputFileName =
                className.replace('.',File.separatorChar)+".class";
            outputFileNames.add(outputFileName);
        }
        
        return outputFileNames;
    }

    
    // -----------------------------------------------------------------------
    // Implementation specific
    // -----------------------------------------------------------------------

    /**
     * Return the class path elements managed by the class loader associated
     * with this instance.
     * 
     * @since 2.2.3
     */
    private List<String> getClassPathEntries() {
        
    	/*
    	 * TODO Investigate whether at some point the .class file could be
    	 * accessed as JavaFileObjects transmitted to compiler getTask. By this
    	 * way, they could be loaded via the class loader of JuliacConfig
    	 * instead of from the class path.
    	 */
    	
        JuliacConfig jconf = jc.getJuliacConfig();
        ClassLoader classLoader = jconf.getClassLoader();
        
        /*
         * I guess that classLoader should never be null (even when recursing
         * into the hierarchy of parent classloaders.) But who knows?
         */
        List<String> lst = new ArrayList<String>();
        while( classLoader!=null && (classLoader instanceof URLClassLoader) ) {
            
            URLClassLoader urlcl = (URLClassLoader) classLoader;
            URL[] urls = urlcl.getURLs();
            for (URL url : urls) {
                String absfilename = url.getPath();
                
                /*
                 * URLs encode special characters such as white spaces.
                 * Remove this encoding.
                 */
                absfilename = absfilename.replace("%20"," ");
                File f = new File(absfilename);
                boolean isRegular = Utils.isRegularClassPathEntry(f);
                if(isRegular) {
                    lst.add(absfilename);
                }
            }
            classLoader = urlcl.getParent();
        }

        return lst;
    }
}
